import enum

class DataObjectState():
    """Basic state of a condition.
    
    Args:    
            - name: String representation of a condition
    """
    
    def __init__(self, name):
        self.name = name
        
    def __repr__(self):
        return "%s" % (self.name)

class ComplexState():    
        def __init__(self, first, last, state_type):
            self.first = first
            self.last = last
            self.type = state_type

class RuntimeState(enum.Enum):
    START = 1
    END = 2

class Type(enum.Enum):
    FOLLOWS = 1
    LEADSTO = 2
    LEADSTOCYCLIC = 3



from .conditions import TimeInstantCondition

def _time_instance_auto_wrap(condition):
    if isinstance(condition, str):
        condition = TimeInstantCondition(condition)
    
    return condition


class _MeasureDefinition():
    def __init__(self):
        super().__init__()
        self.id = None

class CountMeasure(_MeasureDefinition):
    """Measure to calc how many times a condition happened.
    
    Args:
    
            - when: Condition that you want to count
    """

    def __init__(self, when):
        super().__init__()
        self.when = _time_instance_auto_wrap(when)
    
    def __repr__(self):
        return f"CountMeasure ( {self.when} )"

class DataMeasure(_MeasureDefinition):
    """Measure that returns a data from specific filter.
    
    Args:
    
            - data_content_selection: Column you want to select
            - precondition: Filter to apply to data
            - first: True if you want to select first data, False if you want to select last
    """
    
    def __init__(self, data_content_selection, precondition, first):
        super().__init__()
       
        self.data_content_selection = data_content_selection
        self.precondition = _time_instance_auto_wrap(precondition)
        self.first = first


class TimeMeasure(_MeasureDefinition):
    """Measure to calc the time elapsed between A condition and B condition.
    
    Args:
    
            - from_condition: Condition where you want to start counting, defined as A
            - to_condition: End condition, defined as B
            - time_measure_type: Linear in case just want a simple A-B calc, Cyclic in case you want to count 
                all A-B appearances
            - single_instance_agg_function: Only in Cyclic mode. Type of operation applied to our dataset, 
                can be MAX, MIN, SUM, or AVG. 
            - first_to: Only in Linear mode. True to take first appareance of B, false to take last appareance of B
            - precondition: Filter to previusly apply to our dataset
    """

    def __init__(self, from_condition, to_condition, 
                    time_measure_type = 'LINEAR', single_instance_agg_function = 'SUM', 
                         first_to = False, precondition = ''):
        super().__init__()
  
        self.from_condition = _time_instance_auto_wrap(from_condition)
        self.to_condition = _time_instance_auto_wrap(to_condition)
        self.time_measure_type = time_measure_type
        self.single_instance_agg_function = single_instance_agg_function
        self.precondition = precondition
        self.first_to = first_to

    def __repr__(self):
        return f"TimeMeasure ( from={self.from_condition}, to={self.to_condition} )"


class AggregatedMeasure(_MeasureDefinition):
    """Measure to group a dataset by certain time interval.
    
    Args:
    
            - base_measure: First measure applied to the base dataframe.
            - single_instance_agg_function: Type of operation applied to our dataset, 
                can be MAX, MIN, SUM, AVG or GROUPBY.
            - grouper: Array of measures for grouping.
            - filter_to_apply: Measure applied to our base_measure to filter specific values.
    """
    def __init__(self, base_measure, single_instance_agg_function, grouper=None, filter_to_apply=None):
        super().__init__()
  
        self.base_measure = base_measure
        self.filter_to_apply = filter_to_apply
        self.single_instance_agg_function = single_instance_agg_function
        if grouper is None:
            self.grouper = []
        else:
            self.grouper = grouper
        
 
class DerivedMeasure(_MeasureDefinition):
    """Measure that applies an operation to a group of measures.
    
    Args:
    
            - function_expression: Aritmathical or logical expresion as String
            - measure_map: Map of {name_of_measure: Measure}
    """
    
    def __init__(self, function_expression, measure_map):
        super().__init__()
      
        self.function_expression = function_expression
        self.measure_map = measure_map



from ppinot4py.model import *
from ppinot4py.computers import *
import datetime
import pandas as pd
import pytest
import numpy as np

@pytest.fixture
def log_for_time():
    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'
    IdCase3 = '3-364285768'
        
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2011, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2013, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2014, 5, 6, 16, 44, 7)
    time6 = datetime.datetime(2015, 6, 6, 16, 44, 7)
    
        
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase2, IdCase2, IdCase3, IdCase3], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment' ]}

    dataframeLinear = pd.DataFrame(data)

    return dataframeLinear

def test_aggregated_compute_time_grouped(log_for_time):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=True)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA, 
        single_instance_agg_function='SUM')
        
    timeResult1 = datetime.timedelta(days=365, minutes=46, seconds=6)
    
    var = measure_computer(aggregatedMeasure, log_for_time, time_grouper=pd.Grouper(freq='1Y'))
    
    assert var.size == 5
    assert var.iloc[0] == timeResult1
    assert pd.isnull(var.iloc[1])
    assert var.iloc[2] == datetime.timedelta(days=365)

def test_aggregated_compute_time_no_group(log_for_time):

    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=True)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA, 
        single_instance_agg_function='SUM')

    var = measure_computer(aggregatedMeasure, log_for_time)

    assert var == datetime.timedelta(days=1126, minutes=46, seconds =6)

def test_derived_compute_time_condition(log_for_time):
    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=True)
    
    derivedMeasure = DerivedMeasure("time < days366", 
        {"time": timeMeasureLinearA,
         "days366": pd.Timedelta(days=366)})

    var = measure_computer(derivedMeasure, log_for_time)

    assert var.size == 3
    assert var.iloc[0]
    assert var.iloc[1]
    assert not var.iloc[2]

def test_derived_aggregated_time(log_for_time):
    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=True)

    aggregatedMeasure = AggregatedMeasure(
        base_measure=timeMeasureLinearA, 
        single_instance_agg_function='SUM')

    derivedMeasure = DerivedMeasure("time < days366", 
        {"time": aggregatedMeasure,
         "days366": pd.Timedelta(days=366)})

    var = measure_computer(derivedMeasure, log_for_time, time_grouper=pd.Grouper(freq='1Y'))

    assert var.size == 5
    assert np.all(var == [True, False, True, False, False])


def test_count_compute():
    
    countMeasure = CountMeasure('`lifecycle:transition` == "In Progress"')

    IdCase1 = '1-364285768'
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1], 
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress']}
            
    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(countMeasure, dataframeLinear).iloc[0]

    assert var == 2
    
def test_count_compute_not_appear():
    
    countMeasure = CountMeasure('`lifecycle:transition` == "Closed"')
    
    IdCase1 = '1-364285768'
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1], 
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress']}
            
    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(countMeasure, dataframeLinear).iloc[0]

    assert var == 0
    
def test_count_compute_instances():
    
    countMeasure = CountMeasure('`lifecycle:transition` == "In Progress"')
    
    IdCase1 = '1-364285768'
    idCase2 = '2-364285768'
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, idCase2, idCase2, idCase2], 
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'In Progress', 'In Progress', 'In Progress']}
            
    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(countMeasure, dataframeLinear)

    assert var.iloc[0] == 2
    assert var.iloc[1] == 1

def test_data_computer():
    
    dataMeasure = DataMeasure(
        precondition="`concept:name` == 'Queued'", 
        data_content_selection="lifecycle:transition", 
        first=False)

    IdCase1 = '1-364285768'
    
    data = {'case:concept:name': [IdCase1, IdCase1, IdCase1], 
            'concept:name': ['Queued', 'Queued', 'Not queued'],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress']}
            
    dataframeLinear = pd.DataFrame(data)
    result = measure_computer(dataMeasure, dataframeLinear)
    
    assert result.iloc[0] == 'In Progress'

def test_time_linear_instances():
    
    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"', 
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=True)
    
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    
    IdCase1 = '1-364285768'
    
    timeResult = datetime.timedelta(minutes=46, seconds=6) 
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1], 
            'time:timestamp': [time1, time2, time3],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress']}
            
    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(timeMeasureLinearA, dataframeLinear).iloc[0]
    assert var == timeResult

def test_time_linear_instances_with_several_from():
        
    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"', 
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=False)
    
    IdCase1 = '1-364285768'
    
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 6, 16, 44, 7)
    time6 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(minutes=46, seconds=6) 
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'In Progress', 'In Progress', 'In Progress' ]}
    
    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(timeMeasureLinearA, dataframeLinear).iloc[0]
    assert var == timeResult
    
def test_time_linear_instances_WithSeveralToAndFirstTo():
        
    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"', 
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=True)
    
    IdCase1 = '1-364285768'
    
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 6, 16, 44, 7)
    time6 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(minutes=46, seconds=6) 
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'In Progress', 'In Progress', 'In Progress' ]}
    
    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(timeMeasureLinearA, dataframeLinear).iloc[0]
    assert var == timeResult
    

def test_time_linear_instances_WithSeveralToAndNotFirstTo():
    
    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"', 
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=False)
    
    IdCase1 = '1-364285768'
    
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 6, 16, 44, 7)
    time6 = datetime.datetime(2012, 6, 6, 16, 44, 7)

    timeResult = datetime.timedelta(days= 797, hours=23, minutes=44, seconds=25) 
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment' ]}
    
    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(timeMeasureLinearA, dataframeLinear).iloc[0]
    assert var == timeResult
    
    
def test_time_linear_instances_WithSeveralFromAndToAndFirstTo2():
    
    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"', 
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        first_to=False)
    
    IdCase1 = '1-364285768'
    
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)
    
    timeResult = datetime.timedelta(days= 797, hours=23, minutes=44, seconds=25) 
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment']}
    
    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(timeMeasureLinearA, dataframeLinear).iloc[0]
    assert var == timeResult
    
def test_time_cyclic_SumInstances():
    
    
    timeMeasureCyclic = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"', 
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        single_instance_agg_function='SUM',
        time_measure_type='CYCLIC')
    
    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'
    
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)
    
    timeResult = datetime.timedelta(days= 732, minutes=46, seconds=6) 
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase2, IdCase1], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment']}
    
    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(timeMeasureCyclic, dataframeLinear).iloc[0]
    
    assert var == timeResult
    
def test_time_cyclic_MaxInstances():
    
    timeMeasureCyclic = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"', 
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        single_instance_agg_function='MAX',
        time_measure_type='CYCLIC')
    
    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'
    
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)
    
    timeResult = datetime.timedelta(days= 731) 
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase2, IdCase1], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment']}
    
    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(timeMeasureCyclic, dataframeLinear).iloc[0] 
    
    assert var == timeResult
    
def test_time_cyclic_AvgInstances():
    
    timeMeasureCyclic = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"', 
        to_condition='`lifecycle:transition` == "Awaiting Assignment"', 
        single_instance_agg_function='AVG',
        time_measure_type='CYCLIC')
    
    IdCase1 = '1-364285768'
    
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)
    
    timeResult = datetime.timedelta(days=191, hours=12, minutes=11, seconds=31.5) 
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment']}
    
    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(timeMeasureCyclic, dataframeLinear).iloc[0]
    
    assert var == timeResult

def test_derived_instances():
    
    timeMeasureLinearA = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Awaiting Assignment"',
        first_to=True)
    
    timeMeasureLinearB = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Closed"',
        first_to=True)
    
    timeMeasureLinearC = TimeMeasure(
        from_condition='`lifecycle:transition` == "In Progress"',
        to_condition='`lifecycle:transition` == "Pending"',
        first_to=True)
    
    measure_dictionary = {'ProgressTime': timeMeasureLinearA, 'ClosedTime': timeMeasureLinearB, 'PendingTime': timeMeasureLinearC}
    derived_measure = DerivedMeasure(
        function_expression='(ProgressTime + ClosedTime) + PendingTime', 
        measure_map=measure_dictionary)
    
    IdCase1 = '1-364285768'
    
    time_result = datetime.timedelta(days= 3360, minutes=14, seconds=56) 
        
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2011, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2013, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2014, 5, 6, 16, 44, 7)
    time6 = datetime.datetime(2015, 6, 6, 16, 44, 7)
        
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'Closed', 'In Progress', 'Pending']}
    
    dataframeLinear = pd.DataFrame(data)
    var = measure_computer(derived_measure, dataframeLinear)
    
    assert var[0] == time_result


from ppinot4py.computers.conditions_computer import condition_computer
from ppinot4py.model import (
    CountMeasure, 
    DataMeasure, 
    TimeMeasure, 
    AggregatedMeasure,
    DerivedMeasure    
)
from ppinot4py.model.measures import _MeasureDefinition

import pandas as pd
from pandas.api.types import is_datetime64_any_dtype as is_datetime
from pandas.api.types import is_timedelta64_dtype as is_timedelta
from pandas.api.types import is_numeric_dtype, is_bool_dtype
import numpy as np
from copy import copy
import datetime

def measure_computer(measure, dataframe, id_case = 'case:concept:name', time_column = 'time:timestamp', time_grouper = None):
    """ General computer.
    
    Args:    
            - measure: Measure, it will call different computers depending on the type.
            - dataframe: Base dataframe we want to use.
            - id_case (optional): ID column of your dataframe (By default is 'case:concept:name').
            - time_column (optional): Timestamp column of your dataframe (By default is 'time:timestamp').
            - time_grouper (optional): Time grouper (https://pandas.pydata.org/docs/user_guide/timeseries.html)
                without the key. If the measure is aggregated, it groups the result by instance end time

            
    Returns:
            - Series: Series with pairs of case ID - Data

    """
    try:
        # Need to change ":" for "_" because ".query" put nervous with ":"
        #dataframe.columns = [column.replace(":", "_") for column in dataframe.columns]
        # Evaluation wich kind of measure is
        if(type(measure) == CountMeasure):
            computer = count_compute(dataframe,measure, id_case)
        if(type(measure) == DataMeasure):
            computer = data_compute(dataframe,measure, id_case)
        if(type(measure) == TimeMeasure):
            computer = time_compute(dataframe,measure, id_case, time_column)
        if(type(measure) == AggregatedMeasure):
            computer = aggregated_compute(dataframe, measure, id_case, time_column, time_grouper)
        if(type(measure) == DerivedMeasure):
            computer = derived_compute(dataframe, measure, id_case, time_column, time_grouper)
        return computer
    except ValueError as err:
        return f"ERROR: A value in the measure wasn't correctly defined: ({err})"


def count_compute(dataframe, measure, id_case):
    precondition = (measure.when)
    filtered_series = condition_computer(dataframe, id_case, precondition)
    result = filtered_series.groupby(dataframe[id_case]).sum()

    return result

def data_compute(dataframe, measure, id_case):
    precondition = (measure.precondition)
    filtered_series = condition_computer(dataframe, id_case, precondition)
    
    final_dataframe = dataframe[filtered_series]
   
    if(measure.first):
        result = final_dataframe.groupby(id_case)[measure.data_content_selection].first()
    else:
        result = final_dataframe.groupby(id_case)[measure.data_content_selection].last()

    return result

def time_compute(dataframe, measure, id_case, time_column):
    operation = measure.single_instance_agg_function
    precondition = measure.precondition
    time_measure_type = measure.time_measure_type
    from_condition = measure.from_condition
    to_condition = measure.to_condition
    is_first = measure.first_to

    # Initial precondition
    if(precondition != ""):
        filtered_dataframe = dataframe.query(precondition)
    else:
        filtered_dataframe = dataframe
   
    A_condition = condition_computer(filtered_dataframe, id_case, from_condition)
    B_condition = condition_computer(filtered_dataframe, id_case, to_condition)

    removed_timezones = pd.to_datetime(filtered_dataframe[time_column], utc=True)
    dataframe_to_work = pd.DataFrame({'id':filtered_dataframe[id_case], 't': removed_timezones})

    if(time_measure_type == 'LINEAR'): 
        final_result = _linear_time_compute(dataframe_to_work, A_condition, B_condition, is_first, 'id', 't')
   
    elif(time_measure_type == 'CYCLIC'):
        final_result = _cyclic_time_compute(dataframe_to_work, A_condition, B_condition, operation, 'id', 't')
            
    return final_result.reindex(dataframe[id_case].unique())

def _linear_time_compute(dataframeToWork, from_condition, to_condition, is_first, id_case, time_column):
    filtered_dataframe_A = dataframeToWork[from_condition]
    from_values = filtered_dataframe_A.groupby(id_case)[time_column].first()
    
    if(is_first):
        to_values = _first_to_after_from(from_condition, to_condition, dataframeToWork[id_case], dataframeToWork[time_column])
    else:
        finalDataframeB = dataframeToWork[to_condition]        
        to_values = finalDataframeB.groupby(id_case)[time_column].last()
    
    final_result = _linear_calculation(from_values, to_values)

    return final_result

def _first_to_after_from(from_condition, to_condition, id_values, ts_values):
    df = pd.DataFrame({'id':id_values, 't': ts_values})
    df.loc[from_condition, 'c']='A'
    df.loc[to_condition, 'c']='B'

    # Keep only the events with from, to conditions
    df = df.dropna()

    # Choose only those 'to' after a 'from'
    df['prev'] = df.groupby('id')['c'].shift(+1).fillna('B')
    df = df[((df['c']=='B') & (df['prev']=='A'))]

    # Returns the first of them
    return df.groupby('id')['t'].first()

def _linear_calculation(fromValues, toValues):
    
    finalResult = toValues - fromValues   
    finalResult[finalResult < pd.Timedelta(0)] = pd.Timedelta('nan')

#    finalResultWithNaN = finalResultAlmost.groupby(id_case).max()
#    finalResultWithNegatives = finalResultWithNaN.dropna()
#    finalResultDataframe = finalResultWithNegatives.to_frame('data')
        
#    finalResult = finalResultDataframe[finalResultDataframe['data'] > pd.Timedelta(0)]
    #finalResult = finalResultAlmost[finalResultAlmost > pd.Timedelta(0)]
    
    return finalResult    

def _cyclic_time_compute(dataframeToWork, A_condition, B_condition, operation, id_case, time_column):        

    diff = _compute_cyclic_diff(A_condition, B_condition, dataframeToWork[id_case], dataframeToWork[time_column])

    # We remove NaNs so that they do not get evaluated as 0 with sum
    diff = diff.dropna()

    # This is necessary because sum() is not allowed in TimeDeltas
    diff = diff.dt.total_seconds()

    grouped_diff = diff.groupby(dataframeToWork[id_case])

    result = _apply_aggregation(operation, grouped_diff)
        
    return result.apply(lambda x: datetime.timedelta(seconds = x))    

def _apply_aggregation(operation, grouped_df):

    if not isinstance(operation, str):
        raise ValueError('Aggregation function must be a string: sum, min, max, avg or groupby')

    if(operation.upper() == 'SUM'):
        result = grouped_df.sum(min_count=1)

    elif(operation.upper() == "MIN"):
        result = grouped_df.min()

    elif(operation.upper() == "MAX"):
        result = grouped_df.max()

    elif(operation.upper() == "AVG"):
        result = grouped_df.mean()

    elif(operation.upper() == "GROUPBY"):
        result = grouped_df

    else:
        raise ValueError(f'Aggregation operation not valid {operation}. Should be: sum, min, max, avg or groupby')

    return result


def _compute_cyclic_diff(from_condition, to_condition, id_case, timestamps):
    df = pd.DataFrame({'id':id_case, 't': timestamps})
    df.loc[from_condition, 'c']='A'
    df.loc[to_condition, 'c']='B'

    # Keep only the events with from, to conditions
    df = df.dropna()

    # Removes several As (AAA) or several Bs (BBB) by keeping only
    # the first event of each series, i.e., those whose previous 
    # event was different. So, now what we have is a sequence of
    # ABABA or similar
    df['prev'] = df.groupby(id_case)['c'].shift(+1).fillna('B')
    df = df[(((df['c']=='A') & (df['prev']=='B')) | ((df['c']=='B') & (df['prev']=='A')))]

    # Compute the difference for each AB pair
    df_shifted = df.groupby(id_case).shift(-1)
    pair = (df['c']=='A')
    return df_shifted.loc[pair, 't'] - df.loc[pair,'t']

# This method is an alternative implementation to _compute_cyclic_diff
# and might be removed in the future
def _compute_cyclic_diff_alt(from_condition, to_condition, id_case, timestamps):
    combination = _combine(from_condition, to_condition)
    first_change_index = _first_change_index(_detect_change_in_serie(combination, id_case), id_case)
    
    df = pd.DataFrame({'id':id_case, 't': timestamps, 'c': combination})
    df = df.loc[first_change_index]
    df_shifted = df.groupby(id_case).shift(-1)
    pair = (df['c']=='A')
    return df_shifted.loc[pair, 't'] - df.loc[pair,'t']

def _combine(A, B):
    combination = A.copy()
    combination.loc[A] = 'A'
    combination.loc[B] = 'B'
    return combination[(A | B)]

def _detect_change_in_serie(s, id_case):
    return (s != s.groupby(id_case).shift(+1))

def _first_change_index(s, id_case):
    cum_change = s.groupby(id_case).cumsum()
    return s.index.to_series().groupby([id_case, cum_change]).first()

def aggregated_compute(dataframe, measure, id_case, time_column, time_grouper = None):
    base_measure = measure.base_measure
    filter_to_apply = measure.filter_to_apply
    operation = measure.single_instance_agg_function
    data_grouper = measure.grouper
    is_time = False

    base_values = measure_computer(base_measure, dataframe, id_case, time_column)
    
    case_end = dataframe.groupby(id_case)[time_column].last()
    
    if((filter_to_apply is not None) and filter_to_apply != ""):
        filter_condition = measure_computer(filter_to_apply, dataframe, id_case, time_column)
        # We assume the filtered_condition is fine. Maybe we could do
        # some sanity checking here.
        base_values = base_values[filter_condition]
        case_end = case_end[filter_condition]

    # Case end could also be configurable
    internal_df = pd.DataFrame({'data':base_values, 'case_end':case_end})

    if not is_datetime(internal_df['case_end']):
        internal_df['case_end'] = pd.to_datetime(internal_df['case_end'], utc=True)
    
    if is_timedelta(internal_df['data'].dtype):
        internal_df['data'] = internal_df['data'].dt.total_seconds()
        #internal_df['data_seconds'] = internal_df['data_seconds'].fillna(0).astype(float)
        is_time = True

    groupers = []
    if (data_grouper is not None):
        if not isinstance(data_grouper, list):
            data_grouper = [data_grouper]
            
        for gr in data_grouper:
            if not isinstance(gr, str):
                gr_key = gr.id
                if gr_key is None:
                    gr_key = 'group' + str(len(groupers))
            
                internal_df[gr_key] = measure_computer(gr, dataframe, id_case, time_column)
                groupers.append(gr_key)
            else:
                groupers.append(gr)

    if time_grouper is not None:
        internal_time_grouper = copy(time_grouper)
        internal_time_grouper.key = 'case_end'
        groupers.append(internal_time_grouper)

    if len(groupers) > 0:
        result_grouped = internal_df.groupby(groupers)
    else:
        result_grouped = internal_df

    final_result = _apply_aggregation(operation, result_grouped)

    if(operation.upper() == "GROUPBY"):
        is_time = False

    if len(final_result) > 1:
        if is_time == True:
            final_result = final_result['data'].apply(lambda x: datetime.timedelta(seconds = x) if not np.isnan(x) else np.nan)

        final_result = final_result.drop('case_end', axis=0, errors='ignore')
    else:
        if is_time == True:
            final_result = datetime.timedelta(seconds = final_result['data'])
        else:
            final_result = final_result['data']

    return final_result

def derived_compute(dataframe, measure, id_case, time_column, time_grouper=None):
    
    function = measure.function_expression
    measure_map = measure.measure_map
    data_frame_computer = pd.DataFrame()
    istime = False
    
    for key in measure_map: 
        if isinstance(measure_map[key], _MeasureDefinition):
            data_frame_computer[key] = measure_computer(measure_map[key], dataframe, id_case, time_column, time_grouper)        
        else:
            data_frame_computer[key] = measure_map[key]

        if(is_timedelta(data_frame_computer[key].dtype)):
            data_frame_computer[key] = data_frame_computer[key].dt.total_seconds()
            #data_frame_computer[key] = data_frame_computer[key].fillna(0).astype(float)

            # If there is one time, we assume everything is time
            istime = True
  
    evaluated_dataframe = data_frame_computer.eval(function)
    #evaluated_dataframe_noInfinites = evaluated_dataframe.replace([np.inf, -np.inf], 0)
    #final_result = evaluated_dataframe_noInfinites.fillna(0).astype(float)

    final_result = evaluated_dataframe
    
    if istime and is_numeric_dtype(final_result.dtype) and not is_bool_dtype(final_result.dtype):
        final_result = final_result.apply(lambda x: datetime.timedelta(seconds = x) if not np.isnan(x) else np.nan)
    
    return final_result



import datetime
import pandas as pd
from ppinot4py.model import RuntimeState, TimeInstantCondition, AppliesTo
from ppinot4py.computers import condition_computer

def test_time_instant_process_start():
    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'
    
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)
    
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase2, IdCase1], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment']}
    
    dataframeLinear = pd.DataFrame(data)

    cond = TimeInstantCondition(RuntimeState.START, AppliesTo.PROCESS)
    result = condition_computer(dataframeLinear, "case:concept:name", cond)

    result_expected = [True, False, False, False, False, False, True, False]

    assert (result.values == result_expected).all()

def test_time_instant_process_end():
    IdCase1 = '1-364285768'
    IdCase2 = '2-364285768'
    
    time1 = datetime.datetime(2010, 3, 31, 16, 59, 42)
    time2 = datetime.datetime(2010, 3, 31, 17, 45, 48)
    time3 = datetime.datetime(2010, 4, 6, 16, 44, 7)
    time4 = datetime.datetime(2012, 4, 6, 16, 44, 7)
    time5 = datetime.datetime(2012, 5, 1, 16, 44, 7)
    time6 = datetime.datetime(2012, 5, 2, 16, 44, 7)
    time7 = datetime.datetime(2012, 5, 3, 16, 44, 7)
    time8 = datetime.datetime(2012, 6, 6, 16, 44, 7)
    
    
    data = {'case:concept:name':[IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase1, IdCase2, IdCase1], 
            'time:timestamp': [time1, time2, time3, time4, time5, time6, time7, time8],
            'lifecycle:transition': ['In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment', 'In Progress', 'Awaiting Assignment','In Progress', 'Awaiting Assignment']}
    
    dataframeLinear = pd.DataFrame(data)

    cond = TimeInstantCondition(RuntimeState.END, AppliesTo.PROCESS)
    result = condition_computer(dataframeLinear, "case:concept:name", cond)

    result_expected = [False, False, False, False, False, False, True, True]

    assert (result.values == result_expected).all()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ppinot4py", 
    version="1.1.2",
    author="ISA Group",
    author_email="resinas@us.es",
    description="PPINOT for Python (ppinot4py)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/isa-group/ppinot4py",
    packages=setuptools.find_packages(),
    license='GPL 3.0',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'pandas',
        'numpy'
    ]
)


#Possibly not needed
from .states import DataObjectState
import enum

class DataCondition():
    def __init__(self, applies_to):
        self.applies_to = applies_to

    def __repr__(self):
        return "%s" % (self.applies_to)

class AppliesTo(enum.Enum):
    PROCESS = 1
    DATA = 2
class TimeInstantCondition():
    """When condition passes from !condition to condition.
    
    Parameters
    ----------
    changes_to_state : state or str 
        The conditions evaluates to true when there is an event that
        makes this state to become true. If it is a string, it is
        interpreted as a DataObjectState.
    
    applies_to : AppliesTo (default DATA)
        The element of the process to which changes_to_state applies.
    """
    
    def __init__(self, changes_to_state, applies_to=AppliesTo.DATA):
        if isinstance(changes_to_state, str):
            changes_to_state = DataObjectState(changes_to_state)
        
        self.changes_to_state = changes_to_state
        self.applies_to = applies_to
    
    def __repr__(self):
        if self.applies_to == AppliesTo.DATA:
            return f"{self.changes_to_state}"
        else:
            return f"{self.changes_to_state} - {self.applies_to}"




class SeriesCondition():

    def __init__(self, series):
        self.series = series
    
    def __repr__(self):
        return "%s" % (self.series)